// projet2.c
//
// HEROUARD Brice (brice.herouard@utbm.fr)
// COUVIDOU Laurent (laurent.couvidou@utbm.fr)
//
// VALUATION D'EXPRESSIONS BOOLENNES
// Composition d'une expression boolenne :
//  Oprateurs binaires :
//   + = OU
//   . = ET
//  Oprateur unaire :
//   ! = NON
//  Variables :
//   (a,b,...,z)
//   (A,B,...,Z)
//  Constantes :
//   O = FAUX
//   1 = VRAI

#include <stdio.h> // Entres-sorties standard
#include <stdlib.h> // Fonctions standard
#include <string.h> // Traitement des chanes
#include <conio.h> // Affichage sous dos

#define TAI 50 // Taille des chanes et des tableaux


/******************** DECLARATIONS DE TYPES ***********************************/

/****** Type probleme *********************************************************/
typedef int* probleme;

/****** Type expbool **********************************************************/
typedef char expbool[TAI];

/****** Type nature ***********************************************************/
typedef enum{
    var, cte, op_un, op_bin, par_ouv, par_fer, inconnu
}nature;

/****** Type noeud ************************************************************/
typedef struct noeu{
    char valeur;
    struct noeu* filsg; // pointeur vers le fils gauche
    struct noeu* filsd; // pointeur vers le fils droit
}noeud;

/****** Type arbre ************************************************************/
typedef noeud* arbre; // pointeur vers le premier noeud

/****** Type element **********************************************************/
typedef struct elem{
    char valeur;
	struct elem* suivant;
	struct elem* precedent;
}element;

/****** Type liste ************************************************************/
typedef element* liste;


/******************** DECLARATIONS DE FONCTIONS *******************************/


/****** Observateurs de l'arbre binaire ***************************************/

int Vide ( arbre a );
// Teste si a est vide

int Feuille ( arbre a );
// Teste si a est une feuille, c'est--dire si
//  - a est un noeud (soit a non vide)
//  - a n'a ni fils gauche ni fils droit

char Valeur ( arbre a );
// Donne la valeur de la racine de a


/****** Constructeurs de l'arbre binaire **************************************/

arbre CreerFeuille ( char c );
// Cre un arbre d'une feuille contenant le caractre c

arbre Enraciner ( arbre a1, arbre a2, char c );
// Retourne les arbres a1 et a2 lis par une racine contenant c

arbre FilsG ( arbre a );
// Retourne le fils gauche de a

arbre FilsD ( arbre a );
// Retourne le fils droit de a

void Decomposer ( expbool exp, expbool* exp_g, char* op, expbool* exp_d );
// Donne la partie gauche, la partie droite et l'oprateur d'une expression
// Utilise dans ConstruireArbre

arbre ConstruireArbre ( expbool exp, arbre a, probleme pb );
// Construit l'arbre binaire par rcursivit
// La variable pb se charge de signaler les ventuels problmes


/****** Nature d'un caractre *************************************************/

nature DonnerNature ( char lettre );
// Dtermine si un caractre est une variable, une constante, un oprateur...


/****** Construction de la liste **********************************************/

int EstPresent ( liste l, char c );
// Teste la prsence d'un lment c dans une liste l

liste InsererElement ( liste l, char c );
// Insre l'lment c dans la liste l
// Respecte l'ordre alphabtique

liste ConstruireListe ( expbool exp );
// Renvoie une liste construite  partir de l'expression boolenne exp


/****** valuation de l'expression boolenne **********************************/

int Evaluer ( liste l, arbre a, int vals[] );
// value l'expression contenue dans a, pour les vcariables contenues dans l
// prenant leurs valeurs dans vals 

void ConstruireTable ( liste l, arbre a );
// Affiche la table de vrit correspondant  l'expression contenue dans a,
// pour les variables contenues dans l


/****** Divers ****************************************************************/

int Puissance(int n, int x);
// Donne la puissanse n de x


/******************** PROGRAMME PRINCIPAL *************************************/

int main(){

   char encore;
       
   do{
   
       clrscr();
       
       printf("*** Evaluation d'expressions booleennes ***\n\n");
       
       probleme pb = (probleme)malloc(sizeof(int));
       *pb = 0; // La valeur sera change par ConstruireArbre si problme
       
       arbre a;
       a = (noeud*)malloc(sizeof(noeud));
       
       expbool exp; // Reoit l'expression boolenne entre
       
       printf("Entrer l'expression a evaluer : "); // Invite
       scanf("%s",exp);
       getchar(); // "Mange" le caractre retour  la ligne
       
       a = ConstruireArbre ( exp, a, pb ); // Construction de l'arbre
       
       if ( *pb == 0 ){ // Expression correcte
       
           printf("\n\nTable de verite :\n");
	       liste l; // La liste des variables
	 	   l = ConstruireListe(exp);
           ConstruireTable(l , a); // Affichage de la table de vrit
	   }
	   
	   else{ // Expression incorrecte
           printf("\nExpression incorrecte");
           printf("\nExemple correct : ((a.!b+(c.(a+1)))");
       }
  
       printf("\nNouvelle expression (o/n) ? ");
       encore=getchar();
       getchar(); // "Mange" le caractre retour  la ligne
  
  }while ( ( encore == 'o' ) || ( encore == 'O' ) );

  return 0; // Valeur de retour : le programme s'est termin normalement
}


/******************** DEFINITIONS DE FONCTIONS ********************************/


/****** Observateurs de l'arbre binaire ***************************************/

int Vide ( arbre a ){
    return ( a==NULL );
}

int Feuille ( arbre a ){
    return ( (!Vide(a)) && (Vide(FilsG(a))) && (Vide(FilsD(a))) );
}

char Valeur ( arbre a ){
    return a->valeur;
}


/****** Constructeurs de l'arbre binaire **************************************/

arbre CreerFeuille ( char c ){
    arbre nouv;
    nouv = (arbre)malloc(sizeof(noeud));
    nouv->valeur = c;
    nouv->filsg = NULL;
    nouv->filsd = NULL;
    return nouv;
}

arbre Enraciner ( arbre a1, arbre a2, char c ){
    arbre nouv;
    nouv = CreerFeuille(c);
    nouv->filsg = a1;
    nouv->filsd = a2;
    return nouv;
}

arbre FilsG ( arbre a ){
    return a->filsg;
}

arbre FilsD ( arbre a ){
    return a->filsd;
}

void Decomposer ( expbool exp, expbool *exp_g, char *op, expbool *exp_d ){

	int i=1, j=0; // Compteurs
	int position; // Position de l'oprateur
	int longueur=strlen(exp); // Longueur de l'expression
	expbool temp_g, temp_d; // Variables temporaires, pour le traitement
	int nb_par_ouv = 1, nb_op = 0; // Nombre de parenthses et d'oprateurs
	 
	while ( (nb_par_ouv != nb_op) && (i <= longueur) ){
	
		if (DonnerNature(exp[i]) == par_ouv){
            nb_par_ouv++;
        }
		else if (DonnerNature(exp[i]) == op_bin){
            nb_op++;
        }
		i++;
	} // On recherche la position de l'oprateur
	
	position = i-1;
	
	for (i = 0 ; i < position ; i++){
        temp_g[i] = exp[i+1];
    }
	temp_g[position-1] = '\0';
	strcpy(*exp_g,temp_g);
	// Extrait l'expression gauche
	
	for (i = position ; i < longueur - 2 ; i++){
        temp_d[j] = exp[i+1];j++;
    }
	temp_d[j] = '\0';
	strcpy(*exp_d,temp_d);
	// Extrait l'expression droite

	*op = exp[position];
	// Extrait l'oprateur
}

arbre ConstruireArbre ( expbool exp, arbre a, probleme pb ){

    int longueur=strlen(exp); // Longueur de l'expression

	if ( ( (DonnerNature(exp[0]) == var) || (DonnerNature(exp[0]) == cte) ) \
    && (exp[1] == '\0') ){
    
         return Enraciner( NULL, NULL, exp[0]);
    } // Variable ou constante : on cre une feuille
    
	else if ( (DonnerNature(exp[0]) == op_un) ){
	
		int i;
		expbool temp;
		for ( i=0; i<longueur; i++ ){
            temp[i] = exp[i+1];
        }
		temp[i+1] = '\0';
		return Enraciner(NULL,ConstruireArbre(temp,a,pb),'!');
	} // Ngation : un fils droit seul enracin sur l'oprateur
	
    else if (DonnerNature(exp[0]) == par_ouv){
		char temp_op;
		expbool temp_g, temp_d;
		Decomposer (exp, &temp_g, &temp_op, &temp_d );
		return Enraciner ( ConstruireArbre ( temp_g, a, pb ), \
                           ConstruireArbre ( temp_d, a, pb ), \
                           temp_op );
	} // Autres cas : fils droit et fils gauche, enracins sur l'oprateur
					
    *pb = 1; // Problme : un caractre est non conforme
    return NULL; // On renvoie un arbre vide
}


/****** Nature d'un caractre *************************************************/

nature DonnerNature ( char lettre ){

	nature temp = inconnu; // Caractre inconnu au dpart
	
	// On cherche si le carcatre est connu :
	
	if ((lettre >= 'A') && (lettre <= 'z')){ // (A,B,...,Z) et (a,b,...z)
        temp = var;
    }
	if ((lettre == '0') || (lettre == '1')){ // (0 = FAUX) et (1 = VRAI)
        temp = cte;
    }
	if (lettre == '!'){                      // (! = NON)
        temp = op_un;
    }
    if ((lettre == '+') || (lettre == '.')){ // (+ = OU ) et (. = ET)
        temp = op_bin;
    }
	if (lettre == '('){                      // Parenthse ouvrante
        temp = par_ouv;
    }
	if (lettre == ')'){                      // Parenthse fermante
        temp = par_fer;
    }

	return temp;
}


/****** Construction de la liste **********************************************/

int EstPresent ( liste l, char c ){

	int present = 0;

	if (l != NULL){
		while ((l != NULL) && (present == 0)){
			if (l->valeur == c){
                present = 1;
            }
			l = l->suivant;
        }
    }
    
	return present;
}

liste InsererElement ( liste l, char val ){

	liste pt = l; // Pointeur de parcours de la liste
	
	liste temp; // Pointeur temporaire
	temp = (element*)malloc(sizeof(element));
	temp->valeur = val;

	if (pt == NULL){ // La liste est vide : lment unique
	
	    temp->precedent = NULL;
        temp->suivant = NULL;
        l = temp;
    }
    
	else{
	
		while ((pt->suivant != NULL) && (val > pt->valeur)){
            pt = pt->suivant;
        }
        // On avance, tant que le code ASCII du carectre est suprieur 
        // celui rencontr (nb : la table est ASCII dans l'ordre alphabtique)
        
        if (l == pt){ // Liste d'1 lment
        
        	if (val < pt->valeur){ // Le caractre se place avant
                temp->precedent = NULL;
                temp->suivant = pt;
                pt->precedent=temp;
                l=temp;
            }
            
			else{                  // Le caractre se place aprs
                l->suivant = temp;
                temp->precedent = l;
                temp->suivant = NULL;
            }
		}
		
		else if (pt->suivant == NULL){ // Plus grand que tous sauf dernier
		
			if (val < pt->valeur){ // Avant dernier caractre
			
                (pt->precedent)->suivant = temp;
                temp->precedent = pt->precedent;
				temp->suivant = pt;
                pt->precedent= temp;
            }
            
			else{                  // Dernier caractre
			
                pt->suivant = temp;
                temp->precedent = pt;
                temp->suivant = NULL;
            }
		}
		
		else{                      // Cas gnral
		
            (pt->precedent)->suivant = temp;
            temp->precedent = pt->precedent;
			temp->suivant = pt ;
            pt->precedent= temp;
        }
    }
    
	return l;
}

liste ConstruireListe ( expbool exp ){

    int longueur=strlen(exp); // Longueur de l'expression

	int i;
	liste l = NULL;
	  
	for (i = 0 ; i < longueur ; i++){
	    if ( !EstPresent(l,exp[i]) && (DonnerNature(exp[i]) == var) ){
            l = InsererElement( l, exp[i] );
        }
    }
    // On n'insre dans l que les variables, qui n'y sont pas dj prsentes 
    
	return l;
}


/****** valuation de l'expression boolenne **********************************/

int Evaluer ( liste l, arbre a, int vals[] ){

	int i=0; // Compteur
	int eval; // Rsultat de l'expression value : (0 = FAUX) ou (1 = VRAI)

	if ( FilsD(a)==NULL && FilsG(a)==NULL ){ // Variable ou constante
	
		if (DonnerNature(Valeur(a))==var){
            liste tmp = l;
            while (tmp->valeur != Valeur(a)){
                i=i+1;
                tmp=tmp->suivant;
            }
			eval = vals[i]; // On rcupre sa valeur
		}
		
		else if (Valeur(a)=='0'){ // FAUX	
            eval = 0;
        }
        
		else{                     // VRAI
              eval = 1;
     	}
    }
    	
    else if ( Valeur(a)=='.' ){ // "ET" logique
        eval = (Evaluer(l,FilsG(a),vals) && Evaluer(l,FilsD(a),vals));
    }
    
	else if ( Valeur(a)=='+' ){ // "OU" logique
        eval=(Evaluer(l,FilsG(a),vals) || Evaluer(l,FilsD(a),vals));
    }
     
	else{                       // "NON" logique
        eval = !Evaluer(l,FilsD(a),vals);
    }

    return(eval);
}

void ConstruireTable(liste l , arbre a){

	int i, j; // Compteur
    
	int vars[TAI]; // Ligne contenant les valeurs des variables
	int nb_vars = 0; // Nombre de variables
    
    liste pt = l; // Pointeur de parcours de la liste
    
	while ( pt != NULL ){
        nb_vars++;
        pt=pt->suivant;
    } // Compte le nombre d'lments de l ( l'aide de pt)
 
    printf(" ");
    for ( i=1; i<=(2*nb_vars)+4; i++ ){
        printf("-");
    }
    printf("\n|"); // Pour l'affichage

    pt = l;	
    
    while ( pt != NULL ){
        printf(" %c",pt->valeur);
        pt=pt->suivant;
    } // Affiche chaque variable de l (toujours  l'aide de pt)
    
    printf("    |\n|");
    for ( i=1; i<=(2*nb_vars)+4; i++ ){
        printf(" ");
    }
    printf("|\n|"); // Pour l'affichage
   	
    for ( i=0; i<nb_vars; i++ ){
        vars[i]=0;
        printf("% d", vars[i]);
    } // Initialise chaque case de vars  0, et affiche cette 1re ligne
    
    int eval; // Rsultat de l'valuation d'une ligne
    int nb_eval=0; // Nombre d'valuations

	eval = Evaluer ( l , a, vars ); // value la 1re ligne
	printf("  %d |\n",eval); // L'affiche
	
	int ret; // Retenue (cf ci-dessous)

	do{ // Boucle 1 : Traite chaque ligne
	
		vars[nb_vars-1] = vars[nb_vars-1] + 1; // + 1  la ligne en cours
        i=nb_vars-1; // Initialisation du compteur

		do{ // Boucle 2 : Gre les retenues
		
			if ( vars[i]>1 ){          // Si on a dpass 1 dans une variable
				ret=1;                 // Il y a retenue
				vars[i]=0;             // Cette variable passe  0
				vars[i-1]=vars[i-1]+1; // Et la prcdente augmente de 1
				i=i-1;
			}
			else{
			    ret=0;                 // Pas de retenue
            }   
    	}while( (i>=0) && (ret==1) );  // On continue tant qu'il y a retenue
    	
    	printf("|"); // Pour l'affichage
    	
        for( j=0; j<nb_vars; j++ ){
            printf(" %d",vars[j]);
        } // Affiche les valeurs donnes aux variables pour cette ligne
        
		eval = Evaluer ( l , a, vars ); // value la ligne
        printf("  %d |",eval); // Affiche l'valuation
        printf("\n");
        
        nb_eval = nb_eval+1; // On a fait une valuation

    }while ( nb_eval<Puissance(2,nb_vars)-1 );
    // Pour n variables, on aura 2^n valuations  faire, moins une car on a
    // dj valu le cas o toutes les variable sont nulles.
    
    printf(" ");
    for ( i=1; i<=(nb_vars*2)+4; i++ ){
        printf("-");
    }
    printf("\n"); // Pour l'affichage
}


/****** Divers ****************************************************************/

int Puissance(int x, int n){

	int res;

	if (n==0){
        res=1;
	}
    else{
        res=x*Puissance(x,n-1);
	}
	
    return res;
}
